"""Base data provider."""

import contextlib
import json
import operator
import typing as t
from functools import reduce

from mimesis import random as _random
from mimesis.constants import DATADIR, LOCALE_SEP
from mimesis.exceptions import NonEnumerableError
from mimesis.locales import Locale, validate_locale
from mimesis.types import JSON, MissingSeed, Seed

__all__ = ["BaseDataProvider", "BaseProvider"]


class BaseProvider:
    """This is a base class for all providers.


    :attr: random: An instance of :class:`mimesis.random.Random`.
    :attr: seed: Seed for random.
    """

    class Meta:
        name: str

    def __init__(
        self,
        *,
        seed: Seed = MissingSeed,
        random: _random.Random | None = None,
    ) -> None:
        """Initialize attributes.

        Keep in mind that locale-independent data providers will work
        only with keyword-only arguments.

        :param seed: Seed for random.
            When set to `None` the current system time is used.
        :param random: Custom random.
            See https://github.com/lk-geimfari/mimesis/issues/1313 for details.
        """
        if random is not None:
            if not isinstance(random, _random.Random):
                raise TypeError(
                    "The random must be an instance of mimesis.random.Random"
                )
            self.random = random
        else:
            self.random = _random.Random()
        self.seed = seed
        self.reseed(seed)

    def reseed(self, seed: Seed = MissingSeed) -> None:
        """Reseeds the internal random generator.

        In case we use the default seed, we need to create a per instance
        random generator. In this case, two providers with the same seed
        will always return the same values.

        :param seed: Seed for random.
            When set to `None` the current system time is used.
        """
        self.seed = seed
        if seed is MissingSeed:
            # Remove casts after mypy will fix this inference:
            if _random.global_seed is not MissingSeed:
                self.random.seed(t.cast(t.Any, _random.global_seed))
        else:
            self.random.seed(t.cast(t.Any, seed))

    def validate_enum(self, item: t.Any, enum: t.Any) -> t.Any:
        """Validates various enum objects that are used as arguments for methods.

        :param item: Item of an enum object.
        :param enum: Enum object.
        :return: Value of item.
        :raises NonEnumerableError: If enums has not such an item.
        """
        if item is None:
            result = self.random.choice_enum_item(enum)
        elif item and isinstance(item, enum):
            result = item
        else:
            raise NonEnumerableError(enum)

        return result.value

    def _read_global_file(self, file_name: str) -> t.Any:
        """Reads JSON file and return dict.

        Reads JSON file from mimesis/data/global/ directory.

        :param file_name: Path to file.
        :raises FileNotFoundError: If the file was not found.
        :return: JSON data.
        """
        with open(DATADIR.joinpath("global", file_name)) as f:
            data = json.load(f)

        return data

    def _has_seed(self) -> bool:
        """Internal API to check if seed is set."""
        return (self.seed is not None and self.seed is not MissingSeed) or (
            _random.global_seed is not None and _random.global_seed is not MissingSeed
        )

    def __str__(self) -> str:
        """Human-readable representation of locale."""
        return self.__class__.__name__


class BaseDataProvider(BaseProvider):
    """This is a base class for all data providers."""

    def __init__(
        self,
        locale: Locale = Locale.DEFAULT,
        seed: Seed = MissingSeed,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """Initialize attributes for data providers.

        :param locale: Current locale.
        :param seed: Seed to all the random functions.
        """
        super().__init__(seed=seed, *args, **kwargs)
        # This is a dict with data
        # loaded from the JSON file.
        self._dataset: JSON = {}
        # Order matters here, since
        # we have to set up locale first.
        self._setup_locale(locale)
        self._load_dataset()

    def _setup_locale(self, locale: Locale = Locale.DEFAULT) -> None:
        """Set up locale after pre-check.

        :param str locale: Locale
        :raises UnsupportedLocale: When locale not supported.
        :return: Nothing.
        """

        locale_obj = validate_locale(locale)
        self.locale = locale_obj.value

    def _extract(self, keys: list[str], default: t.Any = None) -> t.Any:
        """Extracts nested values from JSON file by list of keys.

        :param keys: List of keys (order extremely matters).
        :param default: Default value.
        :return: Data.
        """
        if not keys:
            raise ValueError("The list of keys to extract cannot be empty.")
        try:
            return reduce(operator.getitem, keys, self._dataset)
        except (TypeError, KeyError):
            return default

    def _update_dict(self, initial: JSON, other: JSON) -> JSON:
        """Recursively updates a dictionary.

        :param initial: Dict to update.
        :param other: Dict to update from.
        :return: Updated dict.
        """
        for k, v in other.items():
            if isinstance(v, dict):
                initial[k] = self._update_dict(initial.get(k, {}), v)
            else:
                initial[k] = other[k]
        return initial

    def _load_dataset(self) -> None:
        """Loads the content from the JSON dataset.

        :return: The content of the file.
        :raises UnsupportedLocale: Raises if locale is unsupported.
        """
        locale = self.locale
        datafile = getattr(self.Meta, "datafile", "")
        datadir = getattr(self.Meta, "datadir", DATADIR)

        if not datafile:
            return None

        def read_file(locale_name: str) -> t.Any:
            file_path = datadir / locale_name / datafile
            with open(file_path, encoding="utf8") as f:
                return json.load(f)

        master_locale = locale.split(LOCALE_SEP).pop(0)

        data = read_file(master_locale)

        if LOCALE_SEP in locale:
            data = self._update_dict(data, read_file(locale))

        self._dataset = data

    def update_dataset(self, data: JSON) -> None:
        """Updates dataset merging a given dict into default data.

        This method may be useful when you need to override data
        for a given key in JSON file.
        """
        if not isinstance(data, dict):
            raise TypeError("The data must be a dict.")

        self._dataset |= data

    def get_current_locale(self) -> str:
        """Returns current locale.

        If locale is not defined, then this method will always return ``en``,
        because ``en`` is default locale for all providers, excluding builtins.

        :return: Current locale.
        """
        # noinspection PyTypeChecker
        return self.locale

    def _override_locale(self, locale: Locale = Locale.DEFAULT) -> None:
        """Overrides current locale with passed and pull data for new locale.

        :param locale: Locale
        :return: Nothing.
        """
        self._setup_locale(locale)
        self._load_dataset()

    @contextlib.contextmanager
    def override_locale(
        self,
        locale: Locale,
    ) -> t.Generator["BaseDataProvider", None, None]:
        """Context manager that allows overriding current locale.

        Temporarily overrides current locale for
        locale-dependent providers.

        :param locale: Locale.
        :return: Provider with overridden locale.
        """
        try:
            origin_locale = Locale(self.locale)
            self._override_locale(locale)
            try:
                yield self
            finally:
                self._override_locale(origin_locale)
        except AttributeError:
            raise ValueError(f"«{self.__class__.__name__}» has not locale dependent")

    def __str__(self) -> str:
        """Human-readable representation of locale."""
        locale = Locale(getattr(self, "locale", Locale.DEFAULT))
        return f"{self.__class__.__name__} <{locale}>"



from typing import Callable

from mimesis.locales import Locale
from mimesis.schema import Field

try:
    import pytest
except ImportError:
    raise ImportError("pytest is required to use this plugin")

_CacheCallable = Callable[[Locale], Field]


@pytest.fixture(scope="session")
def _mimesis_cache() -> _CacheCallable:
    cached_instances: dict[Locale, Field] = {}

    def factory(locale: Locale) -> Field:
        if locale not in cached_instances:
            cached_instances[locale] = Field(locale)
        return cached_instances[locale]

    return factory


@pytest.fixture()
def mimesis_locale() -> Locale:
    """Specifies which locale to use."""
    return Locale.DEFAULT


@pytest.fixture()
def mimesis(_mimesis_cache: _CacheCallable, mimesis_locale: Locale) -> Field:
    """Mimesis fixture to provide fake data using all built-in providers."""
    return _mimesis_cache(mimesis_locale)



"""Provides a random choice from items in a sequence."""
import typing as t

from mimesis.providers.base import BaseProvider

__all__ = ["Choice"]


class Choice(BaseProvider):
    """Class for generating a random choice from items in a sequence."""

    class Meta:
        name = "choice"

    def choice(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """Choice random item form sequence.

        See https://github.com/lk-geimfari/mimesis/issues/619

        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :return: Sequence or uncontained element randomly chosen from items.
        """
        return self.__call__(*args, **kwargs)

    def __call__(
        self,
        items: t.Sequence[t.Any] | None,
        length: int = 0,
        unique: bool = False,
    ) -> t.Sequence[t.Any] | t.Any:
        """Generates a randomly chosen sequence or bare element from a sequence.

        Provide elements randomly chosen from the elements in a sequence
        **items**, where when **length** is specified the random choices are
        contained in a sequence of the same type of length **length**,
        otherwise a single uncontained element is chosen. If **unique** is set
        to True, constrain a returned sequence to contain only unique elements.

        :param items: Non-empty sequence (list, tuple or string) of elements.
        :param length: Length of the sequence (number of elements) to provide.
        :param unique: If True, ensures provided elements are unique.
        :return: Sequence or uncontained element randomly chosen from items.
        :raises TypeError: For non-sequence items or non-integer length.
        :raises ValueError: If negative length or insufficient unique elements.

        >>> from mimesis import Choice
        >>> choice = Choice()

        >>> choice(items=['a', 'b', 'c'])
        'c'
        >>> choice(items=['a', 'b', 'c'], length=1)
        ['a']
        >>> choice(items='abc', length=2)
        'ba'
        >>> choice(items=('a', 'b', 'c'), length=5)
        ('c', 'a', 'a', 'b', 'c')
        >>> choice(items='aabbbccccddddd', length=4, unique=True)
        'cdba'

        """

        if not isinstance(items, t.Sequence):
            raise TypeError("**items** must be non-empty sequence.")

        if not items:
            raise ValueError("**items** must be a non-empty sequence.")

        if length < 0:
            raise ValueError("**length** should be a positive integer.")

        if length == 0:
            return self.random.choice(items)

        if unique and len(set(items)) < length:
            raise ValueError(
                "There are not enough unique elements in "
                "**items** to provide the specified **number**."
            )
        if unique:
            data: list[str] = self.random.sample(list(set(items)), k=length)
        else:
            data = self.random.choices(items, k=length)

        if isinstance(items, list):
            return data
        elif isinstance(items, tuple):
            return tuple(data)
        return "".join(data)



from contextlib import contextmanager
from typing import Any, ClassVar, Iterator

from mimesis.locales import Locale
from mimesis.schema import Field, RegisterableFieldHandlers

try:
    from factory import declarations
    from factory.builder import BuildStep, Resolver
except ImportError:
    raise ImportError("This plugin requires factory_boy to be installed.")

__all__ = ["FactoryField", "MimesisField"]


class FactoryField(declarations.BaseDeclaration):  # type: ignore[misc]
    """
    Mimesis integration with FactoryBoy starts here.

    This class provides a common interface for FactoryBoy,
    but inside it has Mimesis generators.
    """

    _default_locale: ClassVar[Locale] = Locale.EN
    _cached_instances: ClassVar[dict[str, Field]] = {}

    def __init__(
        self,
        field: str,
        locale: Locale | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Creates a field instance.

        The created field is lazy. It also receives build time parameters.
        These parameters are not applied yet.

        :param field: name to be passed to :class:`~mimesis.schema.Field`.
        :param locale: locale to use. This parameter has the highest priority.
        :param kwargs: optional parameters that would be passed to ``Field``.
        """
        super().__init__()
        self.locale = locale
        self.kwargs = kwargs
        self.field = field

    def evaluate(
        self,
        instance: Resolver,
        step: BuildStep,
        extra: dict[str, Any] | None = None,
    ) -> Any:
        """Evaluates the lazy field.

        :param instance: (factory.builder.Resolver): The object holding currently computed attributes.
        :param step: (factory.builder.BuildStep): The object holding the current build step.
        :param extra: Extra call-time added kwargs that would be passed to ``Field``.
        """
        kwargs: dict[str, Any] = {}
        kwargs.update(self.kwargs)
        kwargs.update(extra or {})

        field_handlers = step.builder.factory_meta.declarations.get(
            "field_handlers", []
        )

        _field = self._get_cached_instance(
            locale=self.locale,
            field_handlers=field_handlers,
        )
        return _field(self.field, **kwargs)

    @classmethod
    @contextmanager
    def override_locale(cls, locale: Locale) -> Iterator[None]:
        """
        Overrides unspecified locales.

        Remember that implicit locales would not be overridden.
        """
        old_locale = cls._default_locale
        cls._default_locale = locale
        yield
        cls._default_locale = old_locale

    @classmethod
    def _get_cached_instance(
        cls,
        locale: Locale | None = None,
        field_handlers: RegisterableFieldHandlers | None = None,
    ) -> Field:
        """Returns cached instance.

        :param locale: locale to use.
        :param field_handlers: custom field handlers.
        :return: cached instance of Field.
        """
        if locale is None:
            locale = cls._default_locale

        field_names = "-".join(
            sorted(
                dict(field_handlers if field_handlers else []).keys(),
            )
        )

        key = f"{locale}{field_names}"

        if key not in cls._cached_instances:
            field = Field(locale)

            if field_handlers:
                field.register_handlers(field_handlers)

            cls._cached_instances[key] = field

        return cls._cached_instances[key]


# An alias
MimesisField = FactoryField



"""Specific data provider for Italy (it)."""

import string

from mimesis.enums import Gender
from mimesis.locales import Locale
from mimesis.providers import BaseDataProvider
from mimesis.types import MissingSeed, Seed

__all__ = ["ItalySpecProvider"]


class ItalySpecProvider(BaseDataProvider):
    """Specific-provider of misc data for Italy."""

    def __init__(self, seed: Seed = MissingSeed) -> None:
        """Initialize attributes."""
        super().__init__(locale=Locale.IT, seed=seed)

    class Meta:
        name = "italy_provider"
        datafile = "builtin.json"

    def fiscal_code(self, gender: Gender | None = None) -> str:
        """Return a random fiscal code.

        :param gender: Gender's enum object.
        :return: Fiscal code.

        Example:
            RSSMRA66R05D612U
        """
        code = "".join(self.random.choices(string.ascii_uppercase, k=6))

        code += self.random.generate_string_by_mask(mask="##")

        month_codes = self._extract(["fiscal_code", "month_codes"])
        code += self.random.choice(month_codes)

        birth_day = self.random.randint(101, 131)
        self.validate_enum(gender, Gender)
        if gender == Gender.FEMALE:
            birth_day += 40
        code += str(birth_day)[1:]

        city_letters = self._extract(["fiscal_code", "city_letters"])
        code += self.random.choice(city_letters)
        code += self.random.generate_string_by_mask(mask="###@")

        return code



"""Custom exceptions which are used in Mimesis."""

import typing as t

from mimesis.enums import Locale


class LocaleError(ValueError):
    """Raised when a locale isn't supported."""

    def __init__(self, locale: Locale | str) -> None:
        """Initialize attributes for informative output.

        :param locale: Locale.
        """
        self.locale = locale

    def __str__(self) -> str:
        return f"Invalid locale «{self.locale}»"


class SchemaError(ValueError):
    """Raised when a schema is unsupported."""

    def __str__(self) -> str:
        return (
            "The schema must be a callable object that returns a dict."
            "See https://mimesis.name/en/master/schema.html for more details."
        )


class NonEnumerableError(TypeError):
    """Raised when an object is not an instance of Enum."""

    message = "You should use one item of: «{}» of the object mimesis.enums.{}"

    def __init__(self, enum_obj: t.Any) -> None:
        """Initialize attributes for informative output.

        :param enum_obj: Enum object.
        """
        if enum_obj:
            self.name = enum_obj
            self.items = ", ".join(map(str, enum_obj))
        else:
            self.items = ""

    def __str__(self) -> str:
        return self.message.format(self.items, self.name.__name__)


class FieldError(ValueError):
    """Raised when field is not found."""

    def __init__(self, name: str | None = None) -> None:
        """Initialize attributes for more informative output.

        :param name: Name of the field.
        """
        self.name = name
        self.message = "Field «{}» is not supported."
        self.message_none = "The field cannot have the value None."

    def __str__(self) -> str:
        if self.name is None:
            return self.message_none
        return self.message.format(self.name)


class FieldsetError(ValueError):
    """Raised when a resulting fieldset is empty."""

    def __str__(self) -> str:
        return "The «iterations» parameter should be greater than 1."


class FieldNameError(ValueError):
    """Raised when a field name is invalid."""

    def __init__(self, name: str | None = None) -> None:
        """Initialize attributes for more informative output.

        :param name: Name of the field.
        """
        self.name = name

    def __str__(self) -> str:
        return f"The field name «{self.name}» is not a valid Python identifier."


class FieldArityError(ValueError):
    """Raised when registering field handler has incompatible arity."""

    def __str__(self) -> str:
        return "The custom handler must accept at least two arguments: 'random' and '**kwargs'"


class AliasesTypeError(TypeError):
    """Raised when the aliases attribute is set to a format other than a flat dictionary."""

    def __str__(self) -> str:
        return (
            "The 'aliases' attribute needs to be a non-nested dictionary where "
            "keys are the aliases and values are the corresponding field names."
        )



from pathlib import Path
from typing import Final

__all__ = ["DATADIR", "LOCALE_SEP"]

# This is the path to the data directory in the mimesis package.
DATADIR: Final[Path] = Path(__file__).parent / "datasets"

LOCALE_SEP: Final[str] = "-"



"""Specific data provider for Ukraine (uk)."""

from mimesis.enums import Gender
from mimesis.locales import Locale
from mimesis.providers import BaseDataProvider
from mimesis.types import MissingSeed, Seed

__all__ = ["UkraineSpecProvider"]


class UkraineSpecProvider(BaseDataProvider):
    """Class that provides special data for Ukraine (uk)."""

    def __init__(self, seed: Seed = MissingSeed) -> None:
        """Initialize attributes."""
        super().__init__(locale=Locale.UK, seed=seed)

    class Meta:
        name = "ukraine_provider"
        datafile = "builtin.json"

    def patronymic(self, gender: Gender | None = None) -> str:
        """Generate random patronymic name.

        :param gender: Gender of person.
        :type gender: str or int
        :return: Patronymic name.
        """
        gender = self.validate_enum(gender, Gender)
        patronymics: list[str] = self._extract(["patronymic", str(gender)])
        return self.random.choice(patronymics)
